#include "brainregistration.h"
#include "./src/transformation.cxx"



BrainRegistration::BrainRegistration() {
    rigid = false;
    affine = false;
    bspline = false;

    rigidlevels = 0;
    affinelevels = 0;
    bsplinenodes = 0;

    // Transforms init
    idtransf = IdentityTransformType::New();
    rigidtransf = RigidTransformType::New();
    affinetransf = AffineTransformType::New();
    bsplinetransf = DeformableTransformType::New();
    initializer = TransformInitializerType::New();

    // Optimizer init
    versor_optimizer = VersorOptimizerType::New();
    optimizer = OptimizerType::New();

    // Interpolator init
    interpolator = linInterpolatorType::New();
    binterpolator = BSplineInterpolatorType::New();
    bintinterpolator = BSplineImageInterpolatorType::New();
    bprinterpolator = BSplineAtlasInterpolatorType::New();
    ninterpolator = NNInterpolatorType::New();
    nanainterpolator = NNAnatomicInterpolatorType::New();

    // Metric init
    mi_metric = MIMetricType::New();
    mi_metric->ComputeGradientOff();
    mattesmi_metric = MattesMIMetricType::New();
    mattesmi_metric->ComputeGradientOff();
    mattesmi_metric->SetUseCachingOfBSplineWeights(false);
    mihist_metric = MIHistogramMetricType::New();
    mihist_metric->ComputeGradientOff();
    nmi_metric = NMIMetricType::New();
    nmi_metric->ComputeGradientOff();
    gd_metric = GDMetricType::New();
    gd_metric->ComputeGradientOff();
    kl_metric = KLHistogramMetricType::New();
    kl_metric->ComputeGradientOff();
    kappa_metric = KappaMetricType::New();
    kappa_metric->ComputeGradientOff();
    corr_metric = CorrelationMetricType::New();
    corr_metric->ComputeGradientOff();
    ncorr_metric = NormalizedCorrelationMetricType::New();
    ncorr_metric->ComputeGradientOff();
    ms_metric = MSMetricType::New();
    ms_metric->ComputeGradientOff();
    mshist_metric = MSHistogramMetricType::New();
    mshist_metric->ComputeGradientOff();
    msr_metric = MSReciprocalMetricType::New();
    msr_metric->ComputeGradientOff();

    // Registration method init
    registration = mRegistrationType::New();
    registration->SetMetric(mattesmi_metric);
    registration->SetOptimizer(optimizer);
    registration->SetInterpolator(interpolator);
    multires_registration = MultiResolutionRegistrationType::New();
    multires_registration->SetMetric(mattesmi_metric);
    multires_registration->SetOptimizer(optimizer);
    multires_registration->SetInterpolator(interpolator);

    // Resampling init
    imresample = ResampleImageFilterType::New();
    imresample->SetTransform(this->idtransf);
    imresample->ReleaseDataFlagOn();
    atlasresample = ResampleAtlasFilterType::New();
    atlasresample->SetTransform(this->idtransf);
    atlasresample->ReleaseDataFlagOn();
    maskresample = ResampleMaskFilterType::New();
    maskresample->SetTransform(this->idtransf);
    maskresample->SetInterpolator(ninterpolator);
    maskresample->ReleaseDataFlagOn();
    anatomicresample = ResampleResultsFilterType::New();
    anatomicresample->SetTransform(this->idtransf);
    anatomicresample->SetInterpolator(nanainterpolator);
    anatomicresample->ReleaseDataFlagOn();
    multires_pyramidfix = MultiResolutionPyramidType::New();
    multires_registration->SetFixedImagePyramid(this->multires_pyramidfix);
    multires_pyramidmov = MultiResolutionPyramidType::New();
    multires_registration->SetMovingImagePyramid(this->multires_pyramidmov);

    // Observer init
    observer = ObserverType::New();
    optimizer->AddObserver( itk::IterationEvent(), observer );
    versor_optimizer->AddObserver( itk::IterationEvent(), observer );
    multires_registration->AddObserver( itk::IterationEvent(), observer );

    // Casting init
    fixint2signed = IntensityToSignedCastType::New();
    movint2signed = IntensityToSignedCastType::New();
    transfint2signed = IntensityToSignedCastType::New();

    // Preprocessing init
    matcher = MatchingFilterType::New();

    //Others
    minmax = MinimumMaximumImageCalculatorType::New();
}

void BrainRegistration::SetFixed(ProbabilityImage fix) {

    // Setting fixed image
    this->fiximg=fix;

    // Casting to a signed type for registration
    fixint2signed->SetInput(this->fiximg);
    fixint2signed->Update();

    // Registration parameters
/*    SignedImageType::Pointer down_fiximg = this->DownsampleImage(fixint2signed->GetOutput());

    fixreg = down_fiximg->GetBufferedRegion();
    fixsize = this->fixreg.GetSize();
    origin = down_fiximg->GetOrigin();
    spacing = down_fiximg->GetSpacing();
    griddir = down_fiximg->GetDirection();

    multires_registration->SetFixedImage(down_fiximg);
    registration->SetFixedImage(down_fiximg);
    matcher->SetReferenceImage(down_fiximg);*/
    fixreg = fiximg->GetBufferedRegion();
    fixsize = this->fixreg.GetSize();
    origin = fiximg->GetOrigin();
    spacing = fiximg->GetSpacing();
    griddir = fiximg->GetDirection();

    multires_registration->SetFixedImage(fixint2signed->GetOutput());
    registration->SetFixedImage(fixint2signed->GetOutput());
    matcher->SetReferenceImage(fixint2signed->GetOutput());
    initializer->SetFixedImage(fiximg);

    // Intensity thresholds
    minmax->SetImage(fiximg);
    minmax->Compute();
    mi_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    mattesmi_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    mihist_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    nmi_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    gd_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    kl_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    kappa_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    corr_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    ncorr_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    ms_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    mshist_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);
    msr_metric->SetFixedImageSamplesIntensityThreshold(minmax->GetMaximum()*0.05);


    // No registration done yet
    affine = false;
    rigid = false;
}

void BrainRegistration::SetMoving(ProbabilityImage mov) {
    // Setting moving image
    this->movimg=mov;

    // Retrieving image information
    movreg = this->fiximg->GetBufferedRegion();
    movsize = this->fixreg.GetSize();
    movorigin = this->fiximg->GetOrigin();
    movspacing = this->fiximg->GetSpacing();

    // Casting to a signed type for registration
    movint2signed->SetInput(this->movimg);
    movint2signed->Update();

    /*SignedImageType::Pointer down_movimg = this->DownsampleImage(movint2signed->GetOutput());
    multires_registration->SetMovingImage(down_movimg);
    registration->SetMovingImage(down_movimg);*/
    multires_registration->SetMovingImage(movint2signed->GetOutput());
    registration->SetMovingImage(movint2signed->GetOutput());
    initializer->SetMovingImage(movimg);

    // No registration done yet
    affine = false;
    rigid = false;
}

void BrainRegistration::SetMask(MaskImage mask) {
    this->maskimg=mask;
    this->maskresample->SetInput(this->maskimg);
}

void BrainRegistration::SetAnatomic(ResultsImage anatomic) {
    this->anaimg=anatomic;
    this->anatomicresample->SetInput(this->anaimg);
}

void BrainRegistration::AddAtlas(ProbabilityImage atlas) {
    this->movatlas.push_back(atlas);
}
void BrainRegistration::SetAtlases(std::vector<ProbabilityImage> atlases) {
    this->movatlas.clear();
    for (unsigned int i=0;i<atlases.size();i++) {
        this->movatlas.push_back(atlases[i]);
    }
}

void BrainRegistration::ClearAtlases() {
    this->movatlas.clear();
}

void BrainRegistration::RigidRegistration(int levels, int steps) {
    if (!rigid || rigidlevels < levels) {
        this->bspline=false;
        this->affine=false;

        rigidlevels = levels;

        std::cout << std::endl;
        std::cout << "Initializing rigid transform" << std::endl;
        // Registration initialization
        initializer->SetTransform(rigidtransf);
        initializer->MomentsOn();
        initializer->InitializeTransform();

        // Optimizer parameters
        const double translscale = 1.0 / 1000.0;
        VersorScalesType optimizerscales(rigidtransf->GetNumberOfParameters());
        optimizerscales[0] = 1.0;
        optimizerscales[1] = 1.0;
        optimizerscales[2] = 1.0;
        optimizerscales[3] = translscale;
        optimizerscales[4] = translscale;
        optimizerscales[5] = translscale;
        versor_optimizer->SetScales( optimizerscales );
        versor_optimizer->SetMaximumStepLength( 0.2000  );
        versor_optimizer->SetMinimumStepLength( 0.0001 );
        versor_optimizer->SetNumberOfIterations( steps );

        // Metric parameters
        try {
            mattesmi_metric->SetNumberOfSpatialSamples(static_cast<int>(fixreg.GetNumberOfPixels()*0.5));
        }
        catch( itk::ExceptionObject & err ) {}

        std::cout << "Starting rigid transform" << std::endl;
        // Registration
        multires_registration->SetOptimizer(versor_optimizer);
        multires_registration->SetTransform(rigidtransf);
        multires_registration->SetFixedImageRegion(fixreg);
        multires_registration->SetNumberOfLevels(levels);
        multires_registration->SetInitialTransformParameters(rigidtransf->GetParameters());
        try {
            //registration->StartRegistration();
            multires_registration->StartRegistration();
            this->rigid = true;
        }
        catch( itk::ExceptionObject & err ) {
            std::cerr << "ExceptionObject caught !" << std::endl;
            std::cerr << err << std::endl;
            return;
        }
        rigidtransf->SetParameters(multires_registration->GetLastTransformParameters());

	writeTransformRigid(rigidtransf,"rigidTransform.txt");

        std::cout << "Rigid transform completed" << std::endl;
        std::cout << std::endl;
    }
    imresample->SetTransform(this->rigidtransf);
    atlasresample->SetTransform(this->rigidtransf);
    maskresample->SetTransform(this->rigidtransf);
    anatomicresample->SetTransform(this->rigidtransf);
}

void BrainRegistration::AffineRegistration(int levels, int steps) {
    if (!affine || affinelevels < levels) {
        this->bspline=false;

        affinelevels = levels;

        // Check if images are already rigidly registered
        if (!rigid) {
            this->RigidRegistration();
        }

        std::cout << std::endl;
        std::cout << "Starting affine transform" << std::endl;
        // Transform parameters
        affinetransf->SetCenter(rigidtransf->GetCenter());
        affinetransf->SetTranslation(rigidtransf->GetTranslation());
        affinetransf->SetMatrix(rigidtransf->GetMatrix());

        // Optimizer parameters
        const double translscale = 1.0 / 1000.0;
        OptimizerScalesType optimizerscales(affinetransf->GetNumberOfParameters());
        optimizerscales[0] = 1.0;
        optimizerscales[1] = 1.0;
        optimizerscales[2] = 1.0;
        optimizerscales[3] = 1.0;
        optimizerscales[4] = 1.0;
        optimizerscales[5] = 1.0;
        optimizerscales[6] = 1.0;
        optimizerscales[7] = 1.0;
        optimizerscales[8] = 1.0;
        optimizerscales[9]  = translscale;
        optimizerscales[10] = translscale;
        optimizerscales[11] = translscale;
        optimizer->SetScales( optimizerscales );
        optimizer->SetMaximumStepLength( 0.2000  );
        optimizer->SetMinimumStepLength( 0.0001 );
        optimizer->SetNumberOfIterations( steps );

        // Metric parameters
        try {
            mattesmi_metric->SetNumberOfSpatialSamples(static_cast<int>(fixreg.GetNumberOfPixels()*0.5));
        }
        catch( itk::ExceptionObject & err ) {}

        // Registration
        multires_registration->SetOptimizer(optimizer);
        multires_registration->SetTransform(affinetransf);
        multires_registration->SetNumberOfLevels(levels);
        multires_registration->SetInitialTransformParameters(affinetransf->GetParameters());
        try {
            //registration->StartRegistration();
            multires_registration->StartRegistration();
            affine = true;
        }
        catch( itk::ExceptionObject & err ) {
            std::cerr << "ExceptionObject caught !" << std::endl;
            std::cerr << err << std::endl;
            return;
        }
        affinetransf->SetParameters(multires_registration->GetLastTransformParameters());

	writeTransformAffine(affinetransf,"affineTransform.txt");

        std::cout << "Affine transform completed" << std::endl;
        std::cout << std::endl;
    }
    imresample->SetTransform(this->affinetransf);
    atlasresample->SetTransform(this->affinetransf);
    maskresample->SetTransform(this->affinetransf);
    anatomicresample->SetTransform(this->affinetransf);
}

void BrainRegistration::BSplineMultiRegistration(int levels, unsigned int ngridnodesdim,int steps) {
    if (!bspline || this->bsplinenodes != ngridnodesdim) {
        this->bsplinenodes = ngridnodesdim;

        // Check if images are already affinely registered
        if (!affine) {
            this->AffineRegistration();
        }

        std::cout << std::endl;
        std::cout << "Starting bspline transform" << std::endl;


        // Grid estimation
        gridsizeimg.Fill(ngridnodesdim);
        gridbordersize.Fill(3);    // Border for spline order = 3 ( 1 lower, 2 upper )
        totalgridsize = gridsizeimg + gridbordersize;
        this->bsplinereg.SetSize(totalgridsize);

        for(unsigned int r=0; r<3; r++) {
            spacing[r] *= static_cast<double>(fixsize[r] - 1)  /
                        static_cast<double>(gridsizeimg[r] - 1);
        }

        OriginType gridorig = origin - griddir * spacing;

        // Transform parameters
        bsplinetransf->SetGridSpacing(spacing);
        bsplinetransf->SetGridOrigin(gridorig);
        bsplinetransf->SetGridRegion(this->bsplinereg);
        bsplinetransf->SetGridDirection(griddir);
        bsplinetransf->SetBulkTransform(affinetransf);
        unsigned int nbsplineparam = bsplinetransf->GetNumberOfParameters();
        ParametersType initdeftransfparam(nbsplineparam);
        initdeftransfparam.Fill( 0.0 );
        bsplinetransf->SetParameters(initdeftransfparam);

        // Optimizer parameters
        OptimizerScalesType optimizerscales = OptimizerScalesType(nbsplineparam);
        optimizerscales.Fill( 1.0 );
        optimizer->SetScales( optimizerscales );
        optimizer->SetMaximumStepLength( 10.0 );
        optimizer->SetMinimumStepLength(  0.01 );
        optimizer->SetNumberOfIterations( steps );

        // Metric parameters
        mattesmi_metric->SetUseExplicitPDFDerivatives(false);
        try {
            mattesmi_metric->SetNumberOfSpatialSamples(static_cast<int>(fixreg.GetNumberOfPixels()*0.5));
        }
        catch( itk::ExceptionObject & err ) {}

        // Registration
        multires_registration->SetOptimizer(optimizer);
        multires_registration->SetInterpolator(interpolator);
        multires_registration->SetInitialTransformParameters(bsplinetransf->GetParameters() );
        multires_registration->SetNumberOfLevels(levels);
        multires_registration->SetTransform(bsplinetransf);
        try {
          multires_registration->StartRegistration();
          bspline = true;
        }
        catch( itk::ExceptionObject & err ) {
            std::cerr << "ExceptionObject caught !" << std::endl;
            std::cerr << err << std::endl;
            return;
        }
        bsplinetransf->SetParameters(multires_registration->GetLastTransformParameters());

        writeTransformBSpline(bsplinetransf,"bsplineTransform.txt");

        std::cout << "Bspline transform completed" << std::endl;
        std::cout << std::endl;
    }

    //this->ResampleBSpline();

    imresample->SetInterpolator(bintinterpolator);
    imresample->SetTransform(this->bsplinetransf);
    atlasresample->SetInterpolator(bprinterpolator);
    atlasresample->SetTransform(this->bsplinetransf);
    maskresample->SetTransform(this->bsplinetransf);
    anatomicresample->SetTransform(this->bsplinetransf);
}

void BrainRegistration::BSplineRegistration(unsigned int ngridnodesdim,int steps) {
    if (!bspline || this->bsplinenodes != ngridnodesdim) {
        this->bsplinenodes = ngridnodesdim;

        // Check if images are already affinely registered
        if (!affine) {
            this->AffineRegistration();
        }

        std::cout << std::endl;
        std::cout << "Starting bspline transform" << std::endl;


        // Grid estimation
        gridsizeimg.Fill(ngridnodesdim);
        gridbordersize.Fill(3);    // Border for spline order = 3 ( 1 lower, 2 upper )
        totalgridsize = gridsizeimg + gridbordersize;
        this->bsplinereg.SetSize(totalgridsize);

        for(unsigned int r=0; r<3; r++) {
            spacing[r] *= static_cast<double>(fixsize[r] - 1)  /
                        static_cast<double>(gridsizeimg[r] - 1);
        }

        OriginType gridorig = origin - griddir * spacing;

        // Transform parameters
        bsplinetransf->SetGridSpacing(spacing);
        bsplinetransf->SetGridOrigin(gridorig);
        bsplinetransf->SetGridRegion(this->bsplinereg);
        bsplinetransf->SetGridDirection(griddir);
        bsplinetransf->SetBulkTransform(affinetransf);
        unsigned int nbsplineparam = bsplinetransf->GetNumberOfParameters();
        ParametersType initdeftransfparam(nbsplineparam);
        initdeftransfparam.Fill( 0.0 );
        bsplinetransf->SetParameters(initdeftransfparam);

        // Optimizer parameters
        OptimizerScalesType optimizerscales = OptimizerScalesType(nbsplineparam);
        optimizerscales.Fill( 1.0 );
        optimizer->SetScales( optimizerscales );
        optimizer->SetMaximumStepLength( 10.0 );
        optimizer->SetMinimumStepLength(  0.01 );
        optimizer->SetNumberOfIterations( steps );

        // Metric parameters
        mattesmi_metric->SetNumberOfSpatialSamples(fixreg.GetNumberOfPixels());

        // Registration
        registration->SetOptimizer(optimizer);
        registration->SetInterpolator(interpolator);
        registration->SetInitialTransformParameters(bsplinetransf->GetParameters() );
        registration->SetTransform(bsplinetransf);
        try {
          registration->StartRegistration();
          bspline = true;
        }
        catch( itk::ExceptionObject & err ) {
            std::cerr << "ExceptionObject caught !" << std::endl;
            std::cerr << err << std::endl;
            return;
        }
        bsplinetransf->SetParameters(registration->GetLastTransformParameters());
        std::cout << "Bspline transform completed" << std::endl;
        std::cout << std::endl;
    }


    imresample->SetInterpolator(bintinterpolator);
    imresample->SetTransform(this->bsplinetransf);
    atlasresample->SetInterpolator(bprinterpolator);
    atlasresample->SetTransform(this->bsplinetransf);
    maskresample->SetTransform(this->bsplinetransf);
    anatomicresample->SetTransform(this->bsplinetransf);
}

ProbabilityImage BrainRegistration::GetMoving() {

    // We resample the image to the last transform
    imresample->SetInput(this->movimg);
    imresample->SetSize(this->fiximg->GetLargestPossibleRegion().GetSize());
    imresample->SetOutputOrigin(this->fiximg->GetOrigin());
    imresample->SetOutputSpacing(this->fiximg->GetSpacing());
    imresample->SetOutputDirection(this->fiximg->GetDirection());
    try {
        imresample->Update();
    }
    catch( itk::ExceptionObject & err ) {
        std::cerr << "ExceptionObject caught !" << std::endl;
        std::cerr << err << std::endl;
        return NULL;
    }

    return imresample->GetOutput();

}

MaskImage BrainRegistration::GetMask() {

    // We resample the image to the last transform
    maskresample->SetInput(this->maskimg);
    maskresample->SetSize(this->fiximg->GetLargestPossibleRegion().GetSize());
    maskresample->SetOutputOrigin(this->fiximg->GetOrigin());
    maskresample->SetOutputSpacing(this->fiximg->GetSpacing());
    maskresample->SetOutputDirection(this->fiximg->GetDirection());
    try {
        maskresample->Update();
    }
    catch( itk::ExceptionObject & err ) {
        std::cerr << "ExceptionObject caught !" << std::endl;
        std::cerr << err << std::endl;
        return NULL;
    }

    return maskresample->GetOutput();
}

ResultsImage BrainRegistration::GetAnatomic() {

    // We resample the image to the last transform
    anatomicresample->SetInput(this->anaimg);
    anatomicresample->SetSize(this->fiximg->GetLargestPossibleRegion().GetSize());
    anatomicresample->SetOutputOrigin(this->fiximg->GetOrigin());
    anatomicresample->SetOutputSpacing(this->fiximg->GetSpacing());
    anatomicresample->SetOutputDirection(this->fiximg->GetDirection());
    try {
        anatomicresample->Update();
    }
    catch( itk::ExceptionObject & err ) {
        std::cerr << "ExceptionObject caught !" << std::endl;
        std::cerr << err << std::endl;
        return NULL;
    }

    return anatomicresample->GetOutput();

}

ProbabilityImage BrainRegistration::GetFixed() {
    return(this->fiximg);
}

std::vector<ProbabilityImage> BrainRegistration::GetAtlases() {

    std::vector<ProbabilityImage> atlases;

    // We resample all probabilistic atlases to the last transform
    for (unsigned int i=0;i<this->movatlas.size();i++) {
        atlasresample = ResampleAtlasFilterType::New();
        if (bspline) atlasresample->SetTransform(this->bsplinetransf);
        else if (affine) atlasresample->SetTransform(this->affinetransf);
        else if (rigid) atlasresample->SetTransform(this->rigidtransf);
        else atlasresample->SetTransform(this->idtransf);
        atlasresample->SetSize(this->fiximg->GetLargestPossibleRegion().GetSize());
        atlasresample->SetOutputOrigin(this->fiximg->GetOrigin());
        atlasresample->SetOutputSpacing(this->fiximg->GetSpacing());
        atlasresample->SetOutputDirection(this->fiximg->GetDirection());
        atlasresample->SetInput(this->movatlas[i]);
        try {
            std::cout << "Resampling probabilistic atlas: " << i << std::endl;
            atlasresample->Update();
            atlases.push_back(atlasresample->GetOutput());
        }
        catch( itk::ExceptionObject & err ) {
            std::cerr << "ExceptionObject caught !" << std::endl;
            std::cerr << err << std::endl;
            return atlases;
        }
    }

    return atlases;
}

void BrainRegistration::DemonsRegistration() {
    // We first prepare the histogram matcher
}

void BrainRegistration::UseMutualInformationMetric() {
    registration->SetMetric(this->mi_metric);
    multires_registration->SetMetric(this->mi_metric);
}

void BrainRegistration::UseMattesMutualInformationMetric() {
    registration->SetMetric(this->mattesmi_metric);
    multires_registration->SetMetric(this->mattesmi_metric);
}

void BrainRegistration::UseMutualInformationHistogramMetric() {
    registration->SetMetric(this->mihist_metric);
    multires_registration->SetMetric(this->mihist_metric);
}

void BrainRegistration::UseNormalizedMutualInformationMetric() {
    registration->SetMetric(this->nmi_metric);
    multires_registration->SetMetric(this->nmi_metric);
}

void BrainRegistration::UseGradientDifferenceMetric() {
    registration->SetMetric(this->gd_metric);
    multires_registration->SetMetric(this->gd_metric);
}

void BrainRegistration::UseKullbackLeiblerMetric() {
    registration->SetMetric(this->kl_metric);
    multires_registration->SetMetric(this->kl_metric);
}

void BrainRegistration::UseKappaStatisticMetric() {
    registration->SetMetric(this->kappa_metric);
    multires_registration->SetMetric(this->kappa_metric);
}

void BrainRegistration::UseCorrelationMetric() {
    registration->SetMetric(this->corr_metric);
    multires_registration->SetMetric(this->corr_metric);
}

void BrainRegistration::UseNormalizedCorrelationMetric() {
    registration->SetMetric(this->ncorr_metric);
    multires_registration->SetMetric(this->ncorr_metric);
}

void BrainRegistration::UseMeanSquaresMetric() {
    registration->SetMetric(this->ms_metric);
    multires_registration->SetMetric(this->ms_metric);
}

void BrainRegistration::UseMeanSquaresHistogramMetric() {
    registration->SetMetric(this->mshist_metric);
    multires_registration->SetMetric(this->mshist_metric);
}

void BrainRegistration::UseMeanReciprocalSquaresMetric() {
    registration->SetMetric(this->msr_metric);
    multires_registration->SetMetric(this->msr_metric);
}

void BrainRegistration::ResampleBSpline() {
    /* Resampling B-spline transform */
    unsigned int num_parameters,gridnodesxdir,dim,counter = 0;

    // We double the nodes per grid at each level as the resolution increases and change the B-Splines parameters
    RegionType gridreg;
    RegionType::SizeType gridsize = this->bsplinetransf->GetGridRegion().GetSize();
    gridnodesxdir = gridsize[0]-this->bsplinetransf->SplineOrder;
    gridsize.Fill(gridnodesxdir*2+this->bsplinetransf->SplineOrder);
    gridreg.SetSize(gridsize);

    // We also update the origin and the spacing of the new grid
    SpacingType spacing = fixint2signed->GetOutput()->GetSpacing();
    OriginType  origin  = fixint2signed->GetOutput()->GetOrigin();
    SignedDirectionType griddir = fixint2signed->GetOutput()->GetDirection();
    SignedSizeType fixedsize = fixint2signed->GetOutput()->GetBufferedRegion().GetSize();
    for(dim=0; dim<3; dim++) {
        spacing[dim] *= static_cast<double>(fixedsize[dim] - 1) / static_cast<double>(gridsize[dim] - 1);
    }
    origin -= griddir * spacing;

    // We finally set the number of parameters
    num_parameters = gridsize[0] * gridsize[1] * gridsize[2] * 3;
    ParametersType new_parameters( num_parameters );
    new_parameters.Fill( 0.0 );

    // We create the resampling objects
    ResamplerType::Pointer upsampler = ResamplerType::New();
    FunctionType::Pointer function = FunctionType::New();
    DecompositionType::Pointer decomposition = DecompositionType::New();

    ScaleTransformType::Pointer scale = ScaleTransformType::New();
    ScaleTransformType::ScaleType scale_vector;
    scale_vector.Fill(2.0);
    scale->SetScale(scale_vector);

    for ( dim = 0; dim < 3; dim++ ) {

        // We prepare the grid upsampling
        upsampler->SetInput( this->bsplinetransf->GetCoefficientImage()[dim] );
        upsampler->SetInterpolator( function );
        upsampler->SetTransform( scale );
        upsampler->SetSize( gridsize );
        upsampler->SetOutputSpacing( spacing );
        upsampler->SetOutputOrigin( origin );

        // We prepare the new grid
        decomposition->SetSplineOrder( this->bsplinetransf->SplineOrder );
        decomposition->SetInput( upsampler->GetOutput() );
        decomposition->Update();

        // We copy the new coefficients
        ParametersImageType::Pointer new_coefficients = decomposition->GetOutput();
        Iterator it( new_coefficients, this->bsplinetransf->GetGridRegion() );
        while ( !it.IsAtEnd() ) {
            new_parameters[ counter++ ] = it.Get()*2;
            ++it;
        }
    }

    // We set the final grid parameters
    this->bsplinetransf->SetGridRegion( gridreg );
    this->bsplinetransf->SetGridSpacing( spacing );
    this->bsplinetransf->SetGridOrigin( origin );
    this->bsplinetransf->SetParameters( new_parameters );

}
SignedImageType::Pointer BrainRegistration::DownsampleImage(SignedImageType::Pointer in) {
    // Downsampler
    ResampleSignedImageFilterType::Pointer downsample = ResampleSignedImageFilterType::New();
    downsample->SetTransform(this->idtransf);
    downsample->SetInterpolator( interpolator );
    downsample->SetDefaultPixelValue( 0 );

    // Image info
    SignedImageType::SpacingType inSpacing = in->GetSpacing();

    const double sigmaX = inSpacing[0] * 2;
    const double sigmaY = inSpacing[1] * 2;
    const double sigmaZ = inSpacing[2] * 2;

    // Volume downsampling
    GaussianFilterType::Pointer smootherX = GaussianFilterType::New();
    GaussianFilterType::Pointer smootherY = GaussianFilterType::New();
    GaussianFilterType::Pointer smootherZ = GaussianFilterType::New();

    smootherX->SetInput( in );
    smootherY->SetInput( smootherX->GetOutput() );
    smootherZ->SetInput( smootherY->GetOutput() );
    smootherX->SetSigma( sigmaX );
    smootherY->SetSigma( sigmaY );
    smootherZ->SetSigma( sigmaZ );
    smootherX->SetDirection( 0 );
    smootherY->SetDirection( 1 );
    smootherZ->SetDirection( 2 );
    smootherX->SetNormalizeAcrossScale( false );
    smootherY->SetNormalizeAcrossScale( false );
    smootherZ->SetNormalizeAcrossScale( false );

    SignedImageType::SpacingType outSpacing;
    outSpacing[0] = inSpacing[0] * 2;
    outSpacing[1] = inSpacing[1] * 2;
    outSpacing[2] = inSpacing[2] * 2;
    downsample->SetOutputSpacing( outSpacing );
    downsample->SetOutputOrigin( in->GetOrigin() );
    downsample->SetOutputDirection( in->GetDirection() );

    SignedImageType::SizeType inSize = in->GetBufferedRegion().GetSize();
    inSize[0] = static_cast< SignedSizeValueType >( inSize[0] / 2 );
    inSize[1] = static_cast< SignedSizeValueType >( inSize[1] / 2 );
    inSize[2] = static_cast< SignedSizeValueType >( inSize[2] / 2 );
    downsample->SetSize( inSize );

    downsample->SetInput( smootherZ->GetOutput() );
    downsample->Update();

    return(downsample->GetOutput());
}

ResultsImage BrainRegistration::UpSampling(ResultsImage label, int scale) {
    NNAnatomicInterpolatorType::Pointer interpolator = NNAnatomicInterpolatorType::New();
    ResampleResultsFilterType::Pointer resampler = ResampleResultsFilterType::New();
    IdentityTransformType::Pointer idtransf = IdentityTransformType::New();

    SpacingType spacing = label->GetSpacing();
    OriginType  origin  = label->GetOrigin();
    SignedSizeType size = label->GetBufferedRegion().GetSize();
    SignedDirectionType dir = label->GetDirection();
    size[2] *= scale;
    spacing[2] /= scale;

    resampler->SetTransform(idtransf);
    resampler->SetInterpolator(interpolator);
    resampler->SetInput(label);

    resampler->SetSize( size );
    resampler->SetOutputSpacing( spacing );
    resampler->SetOutputOrigin( origin );

    resampler->Update();

    return(resampler->GetOutput());
}




